/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2022 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.16.1.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : UART implementation

Additional information:

  Device : RX220
  Board  : Renesas RSKRX220

  Unit | UART   | Board connector
  ===============================
  0    | SCI1   | SERIAL

*/

#include "BSP_UART.h"
#include "RTOS.h"  // For OS_INT_Enter()/OS_INT_Leave(). Remove this line and OS_INT_* functions if not using embOS.

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/
#define BSP_UART_CLOCK     (50000000u / 2u)
#define BSP_UART_BAUDRATE  (38400u)

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/
#ifndef EVENACCESS
  #if defined(__RX)
    #define EVENACCESS  __evenaccess
  #else
    #define EVENACCESS
  #endif
#endif

//
// SYSTEM module
//
#define SYSTEM_BASE_ADDR     (0x00080000u)
#define MSTPCRA              (*(volatile EVENACCESS OS_U32*) (SYSTEM_BASE_ADDR + 0x010u))  // Modul stop register A
#define MSTPCRB              (*(volatile EVENACCESS OS_U32*) (SYSTEM_BASE_ADDR + 0x014u))  // Modul stop register B
#define SCKCR                (*(volatile EVENACCESS OS_U32*) (SYSTEM_BASE_ADDR + 0x020u))  // System clock control register
#define SCKCR3               (*(volatile EVENACCESS OS_U16*) (SYSTEM_BASE_ADDR + 0x026u))  // System clock control register 3
#define MOSCWTCR             (*(volatile EVENACCESS OS_U8*)  (SYSTEM_BASE_ADDR + 0x0A2u))  // Main clock oscillator wait control register

//
// Interrupt controller
//
#define IRR_BASE_ADDR        (0x00087000u)
#define IER15                (*(volatile EVENACCESS OS_U8*)  (0x00087215u))
#define IER1B                (*(volatile EVENACCESS OS_U8*)  (0x0008721Bu))
#define IPR218               (*(volatile EVENACCESS OS_U8*)  (0x000873DAu))
#define IPR174               (*(volatile EVENACCESS OS_U8*)  (0x000873AEu))

#define SCI1_BASE_ADDR       (0x0008A020u)

#define PWPR                 (*(volatile EVENACCESS OS_U8*)(0x0008C11Fu))
#define P15PFS               (*(volatile EVENACCESS OS_U8*)(0x0008C14Du))
#define P16PFS               (*(volatile EVENACCESS OS_U8*)(0x0008C14Eu))
#define P26PFS               (*(volatile EVENACCESS OS_U8*)(0x0008C156u))
#define P30PFS               (*(volatile EVENACCESS OS_U8*)(0x0008C158u))

#define P2PMR                (*(volatile EVENACCESS OS_U8*)(0x0008C062u))
#define P3PMR                (*(volatile EVENACCESS OS_U8*)(0x0008C063u))

#define OS_UART_MSTPCR       (MSTPCRB)
#define OS_UART_MSTPCR_BIT   (30u)

#define OS_UART_BASE_ADDR    (SCI1_BASE_ADDR)
#define OS_UART_IPR          (IPR218)
#define OS_UART_IER_ER       (IER1B)
#define OS_UART_IER_RX       (IER1B)
#define OS_UART_IER_TX       (IER1B)
#define OS_UART_IER_ER_BIT   (2u)
#define OS_UART_IER_RX_BIT   (3u)
#define OS_UART_IER_TX_BIT   (4u)
#define OS_UART_ICR          (P2_ICR)
#define OS_UART_ICR_BIT      (1u)

#define OS_UART_ERROR_VECT   214u
#define OS_UART_RX_VECT      215u
#define OS_UART_TX_VECT      216u

#define OS_UART_INT_PRIO     (0x01u)     // Lowest priority

#define OS_UART_SMR          (*(volatile EVENACCESS OS_U8*) (OS_UART_BASE_ADDR + 0x00u))
#define OS_UART_BRR          (*(volatile EVENACCESS OS_U8*) (OS_UART_BASE_ADDR + 0x01u))
#define OS_UART_SCR          (*(volatile EVENACCESS OS_U8*) (OS_UART_BASE_ADDR + 0x02u))
#define OS_UART_TDR          (*(volatile EVENACCESS OS_U8*) (OS_UART_BASE_ADDR + 0x03u))
#define OS_UART_SSR          (*(volatile EVENACCESS OS_U8*) (OS_UART_BASE_ADDR + 0x04u))
#define OS_UART_RDR          (*(volatile EVENACCESS OS_U8*) (OS_UART_BASE_ADDR + 0x05u))
#define OS_UART_SCMR         (*(volatile EVENACCESS OS_U8*) (OS_UART_BASE_ADDR + 0x06u))

#define UART_ERR_PER_BIT     (3u)
#define UART_ERR_FER_BIT     (4u)
#define UART_ERR_ORER_BIT    (5u)
#define UART_ERR_PER_MASK    (1u << UART_ERR_PER_BIT)
#define UART_ERR_FER_MASK    (1u << UART_ERR_FER_BIT)
#define UART_ERR_ORER_MASK   (1u << UART_ERR_ORER_BIT)
#define UART_ERR_MASK_ALL    (UART_ERR_PER_MASK | UART_ERR_FER_MASK | UART_ERR_ORER_MASK)

#define OS_TIMER_FREQ        (20000000u)
#define OS_BAUDRATE          (38400u)
#define OS_BAUDDIVIDE        (((OS_TIMER_FREQ / OS_BAUDRATE) / 32u) - 1u)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static BSP_UART_TX_CB* _pfWriteCB;
static BSP_UART_RX_CB* _pfReadCB;

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _SetBaudrate()
*
*  Function description
*    Configures the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
static void _SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Baudrate);
}

/*********************************************************************
*
*       Global functions, IRQ handler
*
**********************************************************************
*/

/*********************************************************************
*
*       _ISR_RxErr()
*
*  Function description
*    Interrupt handler for UART reception error
*/
#if defined(__ICCRX__)
  #pragma vector = (OS_UART_ERROR_VECT)  // RxError interrupt
  static __interrupt void _ISR_RxErr(void);
  static __interrupt void _ISR_RxErr(void) {
#elif defined(__RX)
  #pragma interrupt (_ISR_RxErr(vect=OS_UART_ERROR_VECT))  // RxError interrupt
  static void _ISR_RxErr(void) {
#elif defined(__GNUC__)
  void __attribute__ ((interrupt)) INT_Excep_SCI1_ERI1(void) {
#endif
  OS_U8 Error;

  OS_EnterNestableInterrupt();
  //
  // Reset Error flags
  //
  Error       = OS_UART_SSR;            // Get current error state
  Error      &= ~(UART_ERR_MASK_ALL);   // Clear all errors
  OS_UART_SSR = Error;                  // Reset errors
  Error       = OS_UART_SSR;            // Dummy read, required according to hardware manual
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       _ISR_Rx()
*
*  Function description
*    Interrupt handler for UART reception
*/
#if defined(__ICCRX__)
  #pragma vector = (OS_UART_RX_VECT)                // Rx data interrupt
  static __interrupt void _ISR_Rx(void);
  static __interrupt void _ISR_Rx(void) {
#elif defined(__RX)
  #pragma interrupt (_ISR_Rx(vect=OS_UART_RX_VECT)) // Rx data interrupt
  static void _ISR_Rx(void) {
#elif defined(__GNUC__)
  void __attribute__ ((interrupt)) INT_Excep_SCI1_RXI1(void) {
#endif
  OS_U8 Data;

  OS_EnterNestableInterrupt();  //  We will enable interrupts
  Data = OS_UART_RDR;
  OS_OnRx(Data);
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       _ISR_Tx()
*
*  Function description
*    Interrupt handler for UART transmission
*/
#if defined(__ICCRX__)
  #pragma vector = (OS_UART_TX_VECT)  // TX data empty interrupt
  static __interrupt void _ISR_Tx(void);
  static __interrupt void _ISR_Tx(void) {
#elif defined(__RX)
  #pragma interrupt (_ISR_Tx(vect=OS_UART_TX_VECT))  // TX data empty interrupt
  static void _ISR_Tx(void) {
#elif defined(__GNUC__)
  void __attribute__ ((interrupt)) INT_Excep_SCI1_TXI1(void) {
#endif
  OS_EnterNestableInterrupt();
  OS_OnTx();
  OS_LeaveNestableInterrupt();
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_UART_Init()
*
*  Function description
*    Initializes the UART peripheral.
*
*  Parameters
*    Unit       : Unit number (typically zero-based).
*    Baudrate   : Baud rate to configure [Hz].
*    NumDataBits: Number of data bits to use.
*    Parity     : One of the following values:
*                   * BSP_UART_PARITY_NONE
*                   * BSP_UART_PARITY_ODD
*                   * BSP_UART_PARITY_EVEN
*    NumStopBits: Number of stop bits to use.
*/
void BSP_UART_Init(unsigned int Unit, unsigned long Baudrate, unsigned char NumDataBits, unsigned char Parity, unsigned char NumStopBits) {
  volatile unsigned int i;
  //
  // Enable UART Module
  //
  OS_UART_MSTPCR &= ~(1uL << OS_UART_MSTPCR_BIT);
  //
  // Set uart port pins
  //
  PWPR   = 0x00u;  // Disable write protect
  PWPR   = 0x40u;  // Enable write access
  P26PFS = 0x0Au;
  P30PFS = 0x0Au;
  PWPR   = 0x80u;  // Enable write protect
  P2PMR  = 0x40u;  // Set pin to peripheral function
  P3PMR  = 0x01u;  // Set pin to peripheral function
  //
  // Reset, disable Reception and transmission
  //
  OS_UART_SCR = 0x00u;
  //
  // Set transmission mode to 8N1, use PCLK
  //
  OS_UART_SMR = 0x00u;
  //
  // Set transfer direction, LSB first
  //
  OS_UART_SCMR = 0x00u;
  //
  // Set baudrate
  //
  OS_UART_BRR = (OS_U8)(OS_BAUDDIVIDE);
  //
  // Wait for SCI Settling time (1 bit cycle)
  //
  for (i = 0u; i < 3000u ; i++) {
  }
  //
  // Clear status
  //
  OS_UART_SSR = 0x00u;
  //
  // Setup interrupt control register for UART
  //
  OS_UART_IPR = OS_UART_INT_PRIO;     // Lowest priority
  //
  // Enable UART interrupts
  //
  OS_UART_IER_ER |= (1u << OS_UART_IER_ER_BIT);  // Enable Rx error interrupt
  OS_UART_IER_RX |= (1u << OS_UART_IER_RX_BIT);  // Enable Rx interrupt
  OS_UART_IER_TX |= (1u << OS_UART_IER_TX_BIT);  // Enable Tx empty interrupt
  //
  // Clear interrupt request flags
  //
  *(OS_U8*)(IRR_BASE_ADDR + OS_UART_ERROR_VECT) = 0u;
  *(OS_U8*)(IRR_BASE_ADDR + OS_UART_RX_VECT)    = 0u;
  *(OS_U8*)(IRR_BASE_ADDR + OS_UART_TX_VECT)    = 0u;
  //
  // Finally enable Rx and Tx interrupts and enable uart Rx and Tx
  //
  OS_UART_SCR |= (1u << 7u)       // Tx interrupt enable
               | (1u << 6u);      // Rx interrupt enable
  OS_UART_SCR |= (1u << 5u)       // Tx enable
               | (1u << 4u);      // Rx enable
}

/*********************************************************************
*
*       BSP_UART_DeInit()
*
*  Function description
*    De-initializes the UART peripheral.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*/
void BSP_UART_DeInit(unsigned int Unit) {
  BSP_UART_USE_PARA(Unit);
}

/*********************************************************************
*
*       BSP_UART_SetBaudrate()
*
*  Function description
*    Configures/changes the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
void BSP_UART_SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  _SetBaudrate(Unit, Baudrate);
}

/*********************************************************************
*
*       BSP_UART_SetReadCallback()
*
*  Function description
*    Sets the callback to execute upon an Rx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetReadCallback(unsigned Unit, BSP_UART_RX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfReadCB = pf;
}

/*********************************************************************
*
*       BSP_UART_SetWriteCallback()
*
*  Function description
*    Sets the callback to execute upon a Tx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetWriteCallback(unsigned int Unit, BSP_UART_TX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfWriteCB = pf;
}

/*********************************************************************
*
*       BSP_UART_Write1()
*
*  Function description
*    Sends one byte via UART.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    Data: (First) data byte to send.
*
*  Additional information
*    The first byte of a transfer is typically sent from application
*    context. Further bytes of the transfer are then sent from the
*    Tx interrupt handler by also calling this function from interrupt
*    context.
*/
void BSP_UART_Write1(unsigned int Unit, unsigned char Data) {
  BSP_UART_USE_PARA(Unit);
  OS_UART_TDR = Data;
}

/*************************** End of file ****************************/
